//
//  HA7LoggerMini03.c
//
//  HONDA GBD-HA7 Vehicle Data Logger Mini 0.3
//  Copyright (C) 2012 Toyohiko Togashi tog001@nifty.com
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//	Description
//		Simple vehicle data logger for "HONDA GBD-HA7".
//		Real time data display to little color LCD.
//		With the record to SD card memory.
//
//	Setting
//		AVR
//			Device	: ATMEL AVR ATmega328P
//			IDE		: AVR Studio 4.18.700  
//			Compiler: WinAVR 20100110(GCC 4.3.3)
//	Usage
//		Plug in an OBD2 socket.
//		It is collecting automatically.
//		Auto power-on and off.
//		Insert SDcard memory, it is starting record.
//		<Special> Push button 3+1 => Reset
//
//		SDcard output file:
//			Media format	: FAT16/12
//			Save directory	: root
//			File name		: "yyyymmdd.CSV"	yyyy=year mm=month dd=days
//			File format		: CSV(camma separated value)
//			   First line is item name
//
//
//	Update history
//	---------- ----- -------------------------------------------
//	2012.03.10 v0.3  First cording
//	2012.03.18 v0.3  Commit
//	2012.07.07 v0.3a Bug fix, RTC initiate
//	2012.07.07 v0.3a Bug fix, stop time reset
//	2012.07.07 v0.3a Lower sensitive a power on condition
//	2012.07.07 v0.3a 20km/l -> 30km/l
//	2012.07.08 v0.3a Update l1f10085.c v0.0 -> v0.1
//	2012.07.08 v0.3a Auto brightness
//	2012.07.11 v0.3a Change of efficiency calculate 
//
//
#define VER_ID	"v0.3a"

#include <stdio.h>						// libc
#include <string.h>						// libc

#include <avr/io.h>						// WinAVR library
#include <avr/interrupt.h>				// WinAVR library
#include <avr/sleep.h>					// WinAVR library
#include <avr/pgmspace.h>				// WinAVR library
#include <util/delay.h>					// WinAVR library

#include "l1f10085.h"					// Private library, LCD module
#include "iso9141_io.h"					// Private library, Vehicle
#include "sd.h"							// Private library, SDcard

#define MAX(a,b)	((a) > (b) ? (a) : (b))
#define MIN(a,b)	((a) < (b) ? (a) : (b))

#define AUTO_BRIGHT						// ADD 2012.07.08  Function enable


//
//	Hardware config
//

// Indicator
#define LED_1_ON()		(PORTD &= ~_BV(PORTD2))				// Neg
#define LED_1_OFF()		(PORTD |=  _BV(PORTD2))
#define LED_1_INIT()	(DDRD  |=  _BV(DDD2), LED_1_OFF())

// Back light
#ifdef AUTO_BRIGHT						// ADD START 2012.07.08
#define LCD_BL_ON()		bl_on()
#define LCD_BL_OFF()	bl_off()
#define LCD_BL_INIT()	bl_init()
#define LCD_BL_TERM()	bl_term()
#define LCD_BL_SET(a)	bl_set(a)
#else									// ADD END 2012.07.08
#define LCD_BL_ON()		(PORTD |=  _BV(PORTD6))
#define LCD_BL_OFF()	(PORTD &= ~_BV(PORTD6))
#define LCD_BL_INIT()	(DDRD  |=  _BV(DDD6), LCD_BL_OFF())
#define LCD_BL_TERM()	(DDRD  &= ~_BV(DDD6), PORTD &= ~_BV(PORTD6))
#define LCD_BL_SET(a)					// ADD 2012.07.08
#endif									// ADD 2012.07.08

// Light senser
#define AMB()			!(PINC & _BV(PINC2))				// Neg
#define AMB_INIT()		((DDRC  &= ~_BV(DDC2)), (PORTC &= ~_BV(PORTC2)))	// Internal pull-ups disable

// Push switch
#define SWITCH_1()		!(PINC & _BV(PINC3))								// Neg
#define SWITCH_1_INIT()	((DDRC  &= ~_BV(DDC3)), (PORTC |= _BV(PORTC3)))		// Internal pull-ups enable
#define SWITCH_2()		!(PINC & _BV(PINC4))								// Neg
#define SWITCH_2_INIT()	((DDRC  &= ~_BV(DDC4)), (PORTC |= _BV(PORTC4)))		// Internal pull-ups enable
#define SWITCH_3()		!(PINC & _BV(PINC5))								// Neg
#define SWITCH_3_INIT()	((DDRC  &= ~_BV(DDC5)), (PORTC |= _BV(PORTC5)))		// Internal pull-ups enable
#define SWITCH_4()		!(PIND & _BV(PIND7))								// Neg
#define SWITCH_4_INIT()	((DDRD  &= ~_BV(DDD7)), (PORTD |= _BV(PORTD7)))		// Internal pull-ups enable

// Beep
#define BEEP_ON()		(PORTD |=  _BV(PORTD5))
#define BEEP_OFF()		(PORTD &= ~_BV(PORTD5))
#define BEEP_INIT()		(DDRD  |=  _BV(DDD5), BEEP_OFF())
#define BEEP_TERM()		(DDRD  &= ~_BV(DDD5), PORTD &= ~_BV(PORTD5))		// Internal pull-ups disable

// SD-WP
#define WP()			(PINB & _BV(PINB0))									//
#define WP_INIT()		((DDRB  &= ~_BV(DDB0)), (PORTB &= ~_BV(PORTB0)))	// Internal pull-ups disable

// SD-DET
#define CD()			!(PINB & _BV(PINB1))								// Neg
#define CD_INIT()		((DDRB  &= ~_BV(DDB1)), (PORTB &= ~_BV(PORTB1)))	// Internal pull-ups disable

// Battery voltage sense register
#define RDIV_1	180						// High side register [kohm]
#define RDIV_2	47						// Low side register [kohm]


//
//	Application config and parameter
//
#define C_VOL 0.000656					// Cylnder volume [m^3]
#define E_CYC 2.0						// Engine output shaft rotation / Cycle
// Function display area
#define FB_CELL_W		19
#define FB_CELL_H		31
#define FB_START_X		0
//#define FB_START_Y	2				// DEL 2012.07.08
#define FB_START_Y		0				// ADD 2012.07.08
#define FB_BORDER_SIZE	1
#define FB_BORDER_COLOR	LCD_COLOR_GREEN
#define FB_CHAR_FCOLOR	LCD_COLOR_GREEN
#define FB_CHAR_SCOLOR	LCD_COLOR_YELLOW
#define FB_CHAR_BCOLOR	LCD_COLOR_BLACK
#define FB_DRAW_CELL(ypos,fcolor,bcolor,str)	\
			lcd_draw_str(\
				(FB_START_X + FB_BORDER_SIZE),\
				((FB_CELL_H + FB_BORDER_SIZE) * (ypos - 1) + FB_BORDER_SIZE + FB_START_Y + 3),\
			    fcolor,\
				bcolor,\
				str)
// Application display area
#define FB_APPL_X		(FB_CELL_W + FB_BORDER_SIZE * 2 + FB_START_X)
#define FB_APPL_Y		0
#define FB_APPL_W		(LCD_SCREEN_WIDTH - FB_APPL_X)
#define FB_APPL_H		(LCD_SCREEN_HEIGHT - FB_APPL_Y)


//
//	Global area
//
double	dist_s   = 0.0;					// Accumulated distance [km]
double	fuse_s   = 0.0;					// Accumulated fuel [l]
long	runtm_s  = 0;					// Accumulated running time [s]
long	ontm_s   = 0;					// Accumulated engine on time [s]
double	dist_t   = 0.0;					// Trip distance [km/day]
double	fuse_t   = 0.0;					// Trip fuel [l/day]
long	runtm_t  = 0;					// Trip running time [s/day]
long	ontm_t   = 0;					// Trip engine on time [s/day]
int		fd       = -1;					//
unsigned char	collect   = 1;			// Enable
unsigned char	log_write = 1;			// Enable


//---------------------------------------------------------------------
//
//	Push key interrupt
//
//---------------------------------------------------------------------
volatile unsigned char	pushkey = 0;

ISR(PCINT1_vect) {
	PCIFR &= ~_BV(PCIF1);

	// Force reboot
	if (SWITCH_1() && SWITCH_3()) {
		asm volatile ("jmp 0x7000");	// Go to bootloader
	}

	if (SWITCH_1()) {
		pushkey = 1;
	}

	if (SWITCH_2()) {
		pushkey = 2;
	}

	if (SWITCH_3()) {
		pushkey = 3;
	}

	return;
}

ISR(PCINT2_vect) {
	PCIFR &= ~_BV(PCIF2);

	if (SWITCH_4()) {
		pushkey = 4;
	}

	return;
}

void pushkey_init(void){
	SWITCH_1_INIT();
	SWITCH_2_INIT();
	SWITCH_3_INIT();
	SWITCH_4_INIT();

	PCMSK1 = _BV(PCINT11)				// PC3(SW1)
	       | _BV(PCINT12)				// PC4(SW2)
	       | _BV(PCINT13)				// PC5(SW3)
		   ;
	PCMSK2 = _BV(PCINT23)				// PD7(SW4)
		   ;
	PCIFR  = 0;							//
	PCICR  = _BV(PCIE1)					// Enable PCINT08-14
	       | _BV(PCIE2);				// Enable PCINT16-23

	return;
}


//---------------------------------------------------------------------
//
//	Real time clock
//
//---------------------------------------------------------------------
// 1 sec interval timer ISR
unsigned char 	ddom[] = {0, 31,28,31,30,31,30,31,31,30,31,30,31};
short 			yy = 2012;
volatile unsigned char 	mm = 7;
volatile unsigned char 	dd = 11;
volatile unsigned char 	hh = 6;
volatile unsigned char 	mi = 0;
volatile unsigned char 	ss = 0;

ISR(TIMER2_OVF_vect) {

	TIFR2 &= ~_BV(TOV2);

	ss++;
	if (ss > 59) {
		ss = 0;
		mi++;
		if (mi > 59) {
			mi = 0;
			hh++;
			if (hh > 23) {
				hh = 0;
				dd++;
				if (((yy & 0x03) == 0) && ((yy % 100) != 0)) {
					ddom[2] = 29;
				} else {
					ddom[2] = 28;
				}
				if (dd > ddom[mm]) {
					dd = 1;
					mm++;
					if (mm > 12) {
						mm = 1;
						yy++;
					}
				}
			}
		}
	}

	return;
}

// Init
void rtc_init(void) {
	//
	// Timer2
	//
										// ADD START 2012.07.07
	TIMSK2 = 0;							// Disable Interrupt
	ASSR =								// Asynchronous Status Register
//			_BV(EXCLK)   |				// Enable External Clock Input
			_BV(AS2)     |				// Asynchronous Timer/Counter2
//			_BV(TCN2UB)  |				// Timer/Counter2 Update Busy
//			_BV(OCR2AUB) |				// Output Compare Register2 Update Busy
//			_BV(OCR2BUB) |				// Output Compare Register2 Update Busy
//			_BV(TCR2AUB) |				// Timer/Counter Control Register2 Update Busy
//			_BV(TCR2BUB) |				// Timer/Counter Control Register2 Update Busy
			0;							//
										// ADD END 2012.07.07
//	DDRB  |=  _BV(DDB3);				// OC2A port output
//	DDRD  |=  _BV(DDD3);				// OC2B port output
	TCCR2A =							// Timer/Counter Control Register A
										// Compare Match Output A Mode
										//   Compare Output, non-PWM Mode
//			_BV(COM2A1) |				//       0  0  1  1
//			_BV(COM2A0) |				//       0  1  0  1
										//       +--|--|--|-- Normal port operation, OC0A disconnected.  <==
										//          +--|--|-- Toggle OC2A on Compare Match
										//             +--|-- Clear OC2A on Compare Match
										//                +-- Set OC2A on Compare Match
										//   Compare Output, Fast PWM Mode
//			_BV(COM2A1) |				//       0  0  1  1
//			_BV(COM2A0) |				//       0  1  0  1
										//       +--|--|--|-- Normal port operation, OC2A disconnected.
										//          +--|--|-- WGM22=0: Normal Port Operation, OC0A Disconnected. WGM22=1: Toggle OC2A on Compare Match.
										//             +--|-- Clear OC2A on Compare Match, set OC2A at BOTTOM,(non-inverting mode).
										//                +-- Set OC2A on Compare Match, clear OC2A at BOTTOM,(invertig mode).
										//   Compare Output, Phase Correct PWM Mode
//			_BV(COM2A1) |				//       0  0  1  1
//			_BV(COM2A0) |				//       0  1  0  1
										//       +--|--|--|-- Normal port operation, OC2A disconnected.
										//          +--|--|-- WGM22=0: Normal Port Operation, OC2A Disconnected. WGM22=1: Toggle OC2A on Compare Match.
										//             +--|-- Clear OC2A on Compare Match when up-counting. Set OC2A on Compare Match when down-counting.
										//                +-- Set OC2A on Compare Match when up-counting. Clear OC2A on Compare Match when down-counting.
										// 
										// Compare Match Output B Mode
										//   Compare Output, non-PWM Mode
//			_BV(COM2B1) |				//       0  0  1  1
//			_BV(COM2B0) |				//       0  1  0  1
										//       +--|--|--|-- Normal port operation, OC2B disconnected.  <==
										//          +--|--|-- Toggle OC2B on Compare Match
										//             +--|-- Clear OC2B on Compare Match
										//                +-- Set OC2B on Compare Match
										//   Compare Output, Fast PWM Mode
//			_BV(COM2B1) |				//       0  0  1  1
//			_BV(COM2B0) |				//       0  1  0  1
										//       +--|--|--|-- Normal port operation, OC2B disconnected.
										//          +--|--|-- Reserved
										//             +--|-- Clear OC2B on Compare Match, set OC2B at BOTTOM, (non-inverting mode).
										//                +-- Set OC2B on Compare Match, clear OC2B at BOTTOM, (inverting mode).
										//   Compare Output, Phase Correct PWM Mode
//			_BV(COM2B1) |				//       0  0  1  1
//			_BV(COM2B0) |				//       0  1  0  1
										//       +--|--|--|-- Normal port operation, OC2B disconnected.
										//          +--|--|-- Reserved
										//             +--|-- Clear OC2B on Compare Match when up-counting. Set OC2B on Compare Match when down-counting.
										//                +-- Set OC2B on Compare Match when up-counting. Clear OC2B on Compare Match when down-counting.
										// Waveform Generation Mode
//			_BV(WGM22) <= TCCR2B		//     0  0  0  0  1  1  1  1
//			_BV(WGM21)  |				//     0  0  1  1  0  0  1  1
//			_BV(WGM20)  |				//     0  1  0  1  0  1  0  1
										//     +--|--|--|--|--|--|--|-- Mode0, Normal  <==
										//        +--|--|--|--|--|--|-- Mode1, PWM, Phase Correct
										//           +--|--|--|--|--|-- Mode2, CTC
										//              +--|--|--|--|-- Mode3, Fast PWM
										//                 +--|--|--|-- Mode4, Reserved
										//                    +--|--|-- Mode5, PWM, Phase Correct
										//                       +--|-- Mode6, Reserved
										//                          +-- Mode7, Fast PWM
			0;							//
	TCCR2B =							// Timer/Counter Control Register B
//			_BV(FOC2A) |				// Force Output Compare A
//			_BV(FOC2B) |				// Force Output Compare B
//			_BV(WGM22) |				// Waveform Generation Mode => See TCCR2A
										// Clock Select
			_BV(CS22)  |				//     0  0  0  0  1  1  1  1
//			_BV(CS21)  |				//     0  0  1  1  0  0  1  1
			_BV(CS20)  |				//     0  1  0  1  0  1  0  1
										//     +--|--|--|--|--|--|--|-- No clock source (Timer/Counter stopped).
										//        +--|--|--|--|--|--|-- clk T2S
										//           +--|--|--|--|--|-- clk T2S/8
										//              +--|--|--|--|-- clk T2S/32
										//                 +--|--|--|-- clk T2S/64
										//                    +--|--|-- clk T2S/128  <==
										//                       +--|-- clk T2S/256
										//                          +-- clk T2S/1024
			0;							//
	TCNT2 =	0;							// Timer/Counter Register
	OCR2A =	0;							// Output Compare Register A
	OCR2B =	0;							// Output Compare Register B
										// ADD START 2012.07.07
	while(ASSR & (						// Wait busy
			_BV(TCN2UB)  |				// Timer/Counter2 Update Busy
			_BV(OCR2AUB) |				// Output Compare Register2 Update Busy
			_BV(OCR2BUB) |				// Output Compare Register2 Update Busy
			_BV(TCR2AUB) |				// Timer/Counter Control Register2 Update Busy
			_BV(TCR2BUB) |				// Timer/Counter Control Register2 Update Busy
			0));						//
										// ADD END 2012.07.07
	TIMSK2 =							// Timer/Counter2 Interrupt Mask Register
//			_BV(OCIE2B) |				// Output Compare Match B Interrupt Enable
//			_BV(OCIE2A) |				// Output Compare Match A Interrupt Enable
			_BV(TOIE2)  |				// Overflow Interrupt Enable
			0;							//
	TIFR2 =								// Timer/Counter2 Interrupt Flag Register
//			_BV(OCF2B)  |				// Output Compare Flag 2 B
//			_BV(OCF2A)  |				// Output Compare Flag 2 A
//			_BV(TOV2)   |				// Overflow Flag
			0;							//
#ifdef DEL								// DEL START 2012.07.07
	ASSR =								// Asynchronous Status Register
//			_BV(EXCLK)   |				// Enable External Clock Input
			_BV(AS2)     |				// Asynchronous Timer/Counter2
//			_BV(TCN2UB)  |				// Timer/Counter2 Update Busy
//			_BV(OCR2AUB) |				// Output Compare Register2 Update Busy
//			_BV(OCR2BUB) |				// Output Compare Register2 Update Busy
//			_BV(TCR2AUB) |				// Timer/Counter Control Register2 Update Busy
//			_BV(TCR2BUB) |				// Timer/Counter Control Register2 Update Busy
			0;							//
	GTCCR |= _BV(PSRASY);				// Prescaler Reset Timer/Counter2 (General Timer/Counter Control Register)
#endif									// DEL END 2012.07.07
	return;
}


//---------------------------------------------------------------------
//
//	A/D converter
//
//---------------------------------------------------------------------
short getadc(unsigned char ch) {
	short	a;

//	DIDR0 = 							// Digital Input Disable Register 0
//		_BV(ADC0D) |					//    ADC0
//		_BV(ADC1D) |					//    ADC1
//		_BV(ADC2D) |					//    ADC2
//		_BV(ADC3D) |					//    ADC3
//		_BV(ADC4D) |					//    ADC4
//		_BV(ADC5D) |					//    ADC5
//		0;
	DIDR0 |= (1 << (ADC0D+ch));
	ADMUX = 							// ADC Multiplexer Selection Register
										// Analog Channel Selection Bits
		(ch << MUX0) |					//    Variable
//		_BV(MUX0) |						//    0  1  0  1  0  1  0  1  0  0  1
//		_BV(MUX1) |						//    0  0  1  1  0  0  1  1  0  1  1
//		_BV(MUX2) |						//    0  0  0  0  1  1  1  1  0  1  1
//		_BV(MUX3) |						//    0  0  0  0  0  0  0  0  1  1  1
										//    ADC0
										//       ADC1
										//          ADC2
										//             ADC3
										//                ADC4
										//                   ADC5
										//                      ADC6
										//                         ADC7
										//                            TempSenser
										//                               VBG=1.1V
										//                                  GND=0V
//		_BV(ADLAR) |					// ADC Left Adjust Result
										// Reference Selection Bits  
		_BV(REFS0) |					//    0  1  0  1
//		_BV(REFS1) |					//    0  0  1  1
										//    AREF, Internal Vref turned off
										//       AVcc with external capacitor at AREF pin
										//          Reserved
										//             Internal 1.1V Voltage Reference with external capacitor at AREF pin
		0;
	ADCSRA = 							// ADC Control and Status Register A
										// ADC Prescaler Select Bits
//		_BV(ADPS0) |					//    0   1   0   1   0    1    0    1
		_BV(ADPS1) |					//    0   0   1   1   0    0    1    1
		_BV(ADPS2) |					//    0   0   0   0   1    1    1    1
										//   1/2 1/2 1/4 1/8 1/16 1/32 1/64 1/128
//		_BV(ADIE)  |					// ADC Interrupt Enable
//		_BV(ADIF)  |					// ADC Interrupt Flag
//		_BV(ADATE) |					// ADC Auto Trigger Enable
		_BV(ADSC)  |					// ADC Start Conversion
		_BV(ADEN)  |					// ADC Enable
		0;

	while((ADCSRA & _BV(ADSC))) ;	// 1/1MHz*16*26cycle=416us
	a = ADCL;
	a |= (ADCH << 8);
	ADCSRA = 0;
	DIDR0 &= ~(1 << (ADC0D+ch));		// Resume I/O port
	return(a);
}


//---------------------------------------------------------------------
//
//	Back light control (Timer/Counter0 8bit PWM)
//
//---------------------------------------------------------------------
#ifdef AUTO_BRIGHT						// ADD START 2012.07.08
void bl_init(void) {
	// Port D6 = OC0A
	DDRD  |=  _BV(DDD6);				// Output
	PORTD &= ~_BV(PORTD6);				// Low level

	// Timer/Counters interrupt
	TIMSK0 = 							// Timer/Counter Interrupt Mask Register
//			_BV(TOIE1)  |				// Timer/Counter1 Overflow Interrupt Enable
//			_BV(OCIE1A) |				// Timer/Counter1 Output Compare A Match Interrupt Enable
//			_BV(OCIE1B) |				// Timer/Counter1 Output Compare B Match Interrupt Enable
//			_BV(ICIE1)  |				// Timer/Counter1 Input Capture Interrupt Enable
//			_BV(OCIE0B) |				// Timer/Counter0 Output Compare Match B Interrupt Enable
//			_BV(TOIE0)  |				// Timer/Counter0 Overflow Interrupt Enable
//			_BV(OCIE0A) |				// Timer/Counter0 Output Compare Match A Interrupt Enable
			0;							//
	TIFR0 = 							// Timer/Counter Interrupt Flag Register
//			_BV(TOV1)   |				// Timer/Counter1 Overflow Flag
//			_BV(OCF1A)  |				// Timer/Counter1 Output Compare A Match Flag
//			_BV(OCF1B)  |				// Timer/Counter1 Output Compare B Match Flag
//			_BV(ICF1)   |				// Timer/Counter1 Input Capture Flag
//			_BV(OCF0B)  |				// Timer/Counter0 Output Compare Flag 0 B
//			_BV(TOV0)   |				// Timer/Counter0 Overflow Flag
//			_BV(OCF0A)  |				// Timer/Counter0 Output Compare Flag 0 A
			0;							//

	// Timer/Counters control register
	TCNT0 = 0;							// Timer/Counter Register
	OCR0A = 0;							// Output Compare Register A  <- Duty 100%
	OCR0B = 0;							// Output Compare Register B  <- Unuse
	bl_on();

	return;
}
void bl_term(void) {
	TCCR0A = 0;							// Reset value
	TCCR0B = 0;							//
	TCNT0  = 0;							//
	OCR0A  = 0;							//
	OCR0B  = 0;							//
	DDRD  &= ~_BV(DDD6);				// Input
	PORTD &= ~_BV(PORTD6);				// No pull-up
}
void bl_on(void) {
	TCCR0A = 							// Timer/Counter0 Control Register A
			_BV(COM0A1) |				// Compare Match Output A Mode  0  0  1  1
			_BV(COM0A0) |				//                              0  1  0  1   non-PWM Mode                               Fast PWM Mode                                        Phase Correct PWM Mode
										//                              +--|--|--|-- Normal port operation, OC0A disconnected   Normal port operation, OC0A disconnected.            Normal port operation, OC0A disconnected
										//                                 +--|--|-- Toggle OC0A on Compare Match               WGM02=0: Normal Port Operation, OC0A Disconnected.   WGM02=0: Normal Port Operation, OC0A Disconnected 
										//                                    |  |                                              WGM02=1: Toggle OC0A on Compare Match.               WGM02=1: Toggle OC0A on Compare Match
										//                                    +--|-- Clear OC0A on Compare Match                Clear OC0A on Compare Match, set OC0A at TOP         Clear OC0A on Compare Match when up-counting. Set OC0A on Compare Match when down-counting
										//                                       +-- Set OC0A on Compare Match                  Set OC0A on Compare Match, clear OC0A at TOP         Set OC0A on Compare Match when up-counting. Clear OC0A on Compare Match when down-counting  <===
//			_BV(COM0B1) |				// Compare Match Output B Mode  0  0  1  1
//			_BV(COM0B0) |				//                              0  1  0  1   non-PWM Mode                               Fast PWM Mode                                        Phase Correct PWM Mode
										//                              +--|--|--|-- Normal port operation, OC0B disconnected   Normal port operation, OC0B disconnected             Normal port operation, OCR0B disconnected
										//                                 +--|--|-- Toggle OC0B on Compare Match               Reserved                                             Reserved
										//                                    +--|-- Clear OC0B on Compare Match                Clear OC0B on Compare Match, set OC0B at TOP         Clear ORC0B on Compare Match when up-counting. Set OCR0B on Compare Match when down-counting
										//                                       +-- Set OC0B on Compare Match                  Set OC0B on Compare Match, clear OC0B at TOP         Set OCR0B on Compare Match when up-counting. Clear OCR0B on Compare Match when down-counting
//		       (WGM02)->TCCR0B			// Waveform Generation Mode  0  0  0  0  1  1  1  1
			_BV(WGM01)  |				//                           0  0  1  1  0  0  1  1                             Update of   TOV Flag 
			_BV(WGM00)  |				//                           0  1  0  1  0  1  0  1   Mode               TOP    OCRx at     Set on
										//                           +--|--|--|--|--|--|--|-- Normal             0xFF   Immediate   MAX
										//                              +--|--|--|--|--|--|-- PWM, Phase Correct 0xFF   TOP         BOTTOM
										//                                 +--|--|--|--|--|-- CTC                OCR0A  Immediate   MAX
										//                                    +--|--|--|--|-- Fast PWM           0xFF   TOP         MAX   <===
										//                                       +--|--|--|-- Reserved            -      -           -
										//                                          +--|--|-- PWM, Phase Correct OCR0A  TOP         BOTTOM
										//                                             +--|-- Reserved            -      -           -
										//                                                +-- Fast PWM           OCR0A  TOP         TOP
			0;							//
	TCCR0B = 							// Timer/Counter0 Control Register B
//			_BV(FOC0A)  |				// Force Output Compare A
//			_BV(FOC0B)  |				// Force Output Compare B
//			_BV(WGM02)  |   			// Waveform Generation Mode -> See TCCR0A
//			_BV(CS02)   |				// Clock Select   0  0  0  0  1  1  1  1
			_BV(CS01)   |				//                0  0  1  1  0  0  1  1
			_BV(CS00)   |				//                0  1  0  1  0  1  0  1
										//                +--|--|--|--|--|--|--|-- No clock source(Timer/Counter stopped)
										//                   +--|--|--|--|--|--|-- clkIO (No prescaling)
										//                      +--|--|--|--|--|-- clkIO/8 (From prescaler)
										//                         +--|--|--|--|-- clkIO/64 (From prescaler)   <=== 8MHz/64/256=488.3Hz
										//                            +--|--|--|-- clkIO/256 (From prescaler)
										//                               +--|--|-- clkIO/1024 (From prescaler)
										//                                  +--|-- External clock source on T0 pin. Clock on falling edge.
										//                                     +-- External clock source on T0 pin. Clock on rising edge.
			0;							//
}
void bl_off(void) {
	TCCR0A = 0;							// Stop counter
	TCCR0B = 0;							//
}
void bl_set(unsigned char bright) {
	OCR0A = ~bright;					// Timer0 PWM duty
}
#endif									// ADD END 2012.07.08


//---------------------------------------------------------------------
//
//	Utilitys
//
//---------------------------------------------------------------------
//
//	Beep sound
//
//void beep_click(void){
//	unsigned char	i;
//
//	for(i = 0; i < 10; i++) {
//		BEEP_ON();
//		_delay_us(500);
//		BEEP_OFF();
//		_delay_us(500);
//	}
//	return;
//}

void beep_approve(void){
	unsigned char	i;

	for(i = 0; i < 100; i++) {
		BEEP_ON();
		_delay_us(500);
		BEEP_OFF();
		_delay_us(500);
	}
	return;
}

//void beep_reject(void){
//	unsigned char	i;
//
//	for(i = 0; i < 100; i++) {
//		BEEP_ON();
//		_delay_us(2000);
//		BEEP_OFF();
//		_delay_us(2000);
//	}
//	return;
//}

void beep_warn(void){
	unsigned char	i;

	for(i = 0; i < 100; i++) {
		BEEP_ON();
		_delay_us(250);
		BEEP_OFF();
		_delay_us(250);
	}
	return;
}

//
// Vertical bar graph
//
void v_gauge(
	unsigned char	x,
	unsigned char 	y,
	unsigned char	w,
	unsigned char 	h,
	unsigned char 	fcolor,
	unsigned char 	fcolor2,
	unsigned char 	lcolor,
	unsigned char 	bcolor,
	float 			value,
	float 			value2,
	float 			max,
	unsigned char 	div)
{
	unsigned char	i;
#define V_GAUGE_A	5
#define V_GAUGE_B	30
#define V_GAUGE_C	60
	unsigned char	x1;
	unsigned char 	y1;
	unsigned char	w1;
	unsigned char 	h1;
	unsigned char 	div1;
	unsigned char	x2;
	unsigned char	y2;
	unsigned char	w2;
	unsigned char	h2;

	// Init
	x1 = x + w * V_GAUGE_A / 100;
	y1 = y + h * V_GAUGE_A / 100;
	w1 = w - w * V_GAUGE_A * 2 / 100;
	h1 = h - h * V_GAUGE_A * 2 / 100;
	div1 = div + 1;

	// Clear black
	lcd_draw_rect(x, y, w, h, bcolor);

	// Scale line and caption
	for(i = 0; i < div1; i++) {
		unsigned char	b;
		char			buff[16];

		b = h1 * i / div + y1;
		lcd_draw_line(x1, b, x1+w1, b, 1, lcolor);

		sprintf(buff, "%d", (int)((div - i) * max / div));
		lcd_draw_str(x1, b-3, lcolor, 0, buff);
	}

	// Bar box
	x2 = w1 * V_GAUGE_B / 100 + x1;
	y2 = (max - value) * h1 / max + y1;
	w2 = w1 * V_GAUGE_C / 100;
	h2 = value * h1 / max;
	lcd_draw_rect(x2, y2, w2, h2, fcolor);

	// Bar line
	y2 = (max - value2) * h1 / max + y1;
	lcd_draw_rect(x2, y2-1, w2, 3, fcolor2);

	return;
}

//
// Horizontal bar graph
//
void h_gauge(
	unsigned char	x,
	unsigned char 	y,
	unsigned char	w,
	unsigned char 	h,
	unsigned char 	fcolor,
	unsigned char 	fcolor2,
	unsigned char 	lcolor,
	unsigned char 	bcolor,
	float 			value,
	float 			value2,
	float 			max,
	unsigned char 	div)
{
	unsigned char	i;
#define H_GAUGE_A	5
#define H_GAUGE_B	10
#define H_GAUGE_C	60
	unsigned char	x1;
	unsigned char 	y1;
	unsigned char	w1;
	unsigned char 	h1;
	unsigned char 	div1;
	unsigned char	x2;
	unsigned char	y2;
	unsigned char	w2;
	unsigned char	h2;

	// Init
	x1 = x + w * H_GAUGE_A / 100;
	y1 = y + h * H_GAUGE_A / 100;
	w1 = w - w * H_GAUGE_A * 2 / 100;
	h1 = h - h * H_GAUGE_A * 2 / 100;
	div1 = div + 1;

	// Clear black
	lcd_draw_rect(x, y, w, h, bcolor);

	// Scale line and caption
	for(i = 0; i < div1; i++) {
		unsigned char	b;
		char			buff[16];

		b = w1 * i / div + x1;
		lcd_draw_line(b, y1, b, y1+h1, 1, lcolor);

		sprintf(buff, "%d", (int)(i * max / div));
		lcd_draw_str(b-3, y1+h1-7, lcolor, 0, buff);
	}

	// Bar box
	x2 = x1;
	y2 = h1 * H_GAUGE_B / 100 + y1;
	w2 = value * w1 / max;
	h2 = h1 * H_GAUGE_C / 100;
	lcd_draw_rect(x2, y2, w2, h2, fcolor);

	// Bar line
	x2 = value2 * w1 / max + x1;
	lcd_draw_rect(x2-1, y2, 3, h2, fcolor2);

	return;
}

//
// Redirect stdout -> SDcard
//
int sd_putchar(char c, FILE *stream) {
	sd_write(fd, &c, 1);
	return 0;
}
FILE lcd_iob = FDEV_SETUP_STREAM(sd_putchar, NULL, _FDEV_SETUP_WRITE);

//
//	Setup
//
void setup(void) {
#define SETUP_ITEM	8
	const char *fb_msg2[] = {
		"-",
		" ^ \nU P\n   ",
		"   \nSET\n   ",
		"   \nDWN\n v ",
		"   \nEnd\n   "
	};
	unsigned char	i;
	unsigned char	x1;
	unsigned char	y1;
	unsigned char	s;			// Selected item
	unsigned char	v;			// Change value
	char			msg[32];

	// Function display
	for(i = 1; i <= 4; i++) {
		FB_DRAW_CELL(i, FB_CHAR_FCOLOR, FB_CHAR_BCOLOR, fb_msg2[i]);
	}
	lcd_draw_rect(FB_APPL_X, FB_APPL_Y, FB_APPL_W, FB_APPL_H, LCD_COLOR_BLACK);

	// Enter setup mode
	s = 0;
	v = 0;
	for(;;) {

		// Display values
		x1 = FB_APPL_X + 4;
		y1 = FB_APPL_X + 4;
		for(i = 0; i < SETUP_ITEM; i++) {
			unsigned char	c1;
			unsigned char	c2;

			if (i == s) {
				if (v) {
					c1 = LCD_COLOR_BLACK;
					c2 = LCD_COLOR_YELLOW;
				} else {
					c1 = LCD_COLOR_YELLOW;
					c2 = LCD_COLOR_BLACK;
				}
			} else {
				c1 = LCD_COLOR_GREEN;
				c2 = LCD_COLOR_BLACK;
			}

			switch(i) {
			case 0:
				sprintf_P(msg, PSTR("Year   :%4d"), yy);
				break;
			case 1:
				sprintf_P(msg, PSTR("Mont   :%4d"), mm);
				break;
			case 2:
				sprintf_P(msg, PSTR("Day    :%4d"), dd);
				break;
			case 3:
				sprintf_P(msg, PSTR("Hour   :%4d"), hh);
				break;
			case 4:
				sprintf_P(msg, PSTR("Min    :%4d"), mi);
				break;
			case 5:
				sprintf_P(msg, PSTR("Sec    :%4d"), ss);
				break;
			case 6:
				sprintf_P(msg, PSTR("Connect: %s"), collect?"ON ":"OFF");
				break;
			case 7:
				sprintf_P(msg, PSTR("Log    : %s"), log_write?"ON ":"OFF");
				break;
			default:
				break;
			}
			lcd_draw_str(x1, y1, c1, c2, msg);
			y1 += 10;

		}

		// Push keys action
		while(SWITCH_1() || SWITCH_2() || SWITCH_3() || SWITCH_4());
		while(!SWITCH_1() && !SWITCH_2() && !SWITCH_3() && !SWITCH_4());

		if (SWITCH_1()) {		// Up
			if (!v) {
				if (s == 0) {
					s = SETUP_ITEM;
				}
				--s;
			} else {
				switch(s){
				case 0:	yy++; break;
				case 1:	mm++; break;
				case 2:	dd++; break;
				case 3:	hh++; break;
				case 4:	mi++; break;
				case 5:	ss = 0; break;
				case 6:	collect = !collect; break;
				case 7:	log_write = !log_write; break;
				default: break;
				}
			}
		} else if (SWITCH_2()) {// Set
			v = !v;
		} else if (SWITCH_3()) {// Down
			if (!v) {
				s++;
				if (s >= SETUP_ITEM) {
					s = 0;
				}
			} else {
				switch(s){
				case 0:	--yy; break;
				case 1:	--mm; break;
				case 2:	--dd; break;
				case 3:	--hh; break;
				case 4:	--mi; break;
				case 5:	ss = 0; break;
				case 6:	collect = !collect; break;
				case 7:	log_write = !log_write; break;
				default: break;
				}
			}
		} else {				// End
			if (v) {
				v = 0;
			} else {
				break;
			}
		}

		// Anti chattaring
		_delay_ms(200);
	}

	return;
}


//---------------------------------------------------------------------
//
// Main
//
//---------------------------------------------------------------------
int main(void){
	unsigned char	disp_mode;			//
	unsigned char	disp_mode_old;		//
	unsigned char	line_init;			// Connect flag
	unsigned char	time_out;			// Connect error count
	unsigned char	ss_old;				//
	unsigned char	dd_old;				//
	unsigned char	pushkey_c;			// Long push counter
	short			stoptm;				// Stopping time of engine on [s]

	const char *fb_msg[] = {
		"-",
		"E  \nffc\n   ",
		"S  \nped\n   ",
		"G  \near\n   ",
		"Sub\npnl\n   ",
		"E  \nffc\nVal",
		"De \ntal\n   ",
		"   \n - \n   ",
		"Set\n up\n   "
	};


	//---------------------------------------------------------------------
	//	Init
	//---------------------------------------------------------------------

	// Master CPU clock
	CLKPR = _BV(CLKPCE);				// Change internal Clock divide
	CLKPR = 0;							// defult 1MHz -> 8MHz
	set_sleep_mode(SLEEP_MODE_PWR_SAVE);	// Only clkASY & TimerOSC

	// Basic I/O
	LED_1_INIT();
	BEEP_INIT();
	CD_INIT();
	WP_INIT();
	pushkey_init();
	rtc_init();
	sei();								// Global interrupt

	// Display
	LCD_BL_INIT();
	LCD_BL_ON();
	lcd_open();
	lcd_draw_str(28, 60, LCD_COLOR_WHITE, LCD_COLOR_BLACK, "HA7 logger mini");
	lcd_draw_str(28, 70, LCD_COLOR_WHITE, LCD_COLOR_BLACK, VER_ID);

	// Communication channel
	iso9141_open();

	// Application init
	disp_mode = 1;						// Default setup mode
	disp_mode_old = 0;					// 
	line_init = 0;
	time_out  = 0;
	stdout    = &lcd_iob;
	ss_old    = 0;
	dd_old    = 0;
	pushkey_c = 0;
	stoptm    = 0;


	//---------------------------------------------------------------------
	//	Main loop
	//---------------------------------------------------------------------
	for(;;){
		short			ect;			// Engine Coolant Temperature -40..215["C]
		signed char		shrtft1;		// Short Term Fuel Trim - Bank 1 -100..99[%]
		signed char		longft1;		// Long Term Fuel Trim - Bank 1 -100..99[%]
		signed char		map;			// Intake Manifold Absolute Pressure 0..255[kPa]
		short			rpm;			// Engine RPM 0..16383[min^-1]
		unsigned char	vss;			// Vehicle Speed Sensor 0..255[km/h] 
		signed char		sparkadv;		// Ignition Timing Advance for #1 Cylinder -64..63["]
		short			iat;			// Intake Air Temperature -40..215["C]
		unsigned char	tp;				// Absolute Throttle Position 0..100[%]
		short			o2s11;			// Bank 1 Sensor 1 - Oxygen Sensor Output Voltage 0..1275[mV]
		signed char		shrtft11;		// Bank 1 Sensor 1 - Short Term Fuel Trim -100..99[%]
		unsigned short	vpwr;			// Control module voltage 0..65535[mV]

		unsigned char	intvl;			// Interval of measure [s]
		float			maf;			// Air Flow Rate from Mass Air Flow Senser [g/s]
		float			mff;			// Fuel flow on intake manifold [g/s]
		float			r;				// Air density [kg/m^3]
		float			v;				// Air volume [m^3/sec]
		float			atm;			// Air pressure [atm]
		float			dist;			// Distance [km]
		float			fuse;			// Used fuel [l]
		float			eff;			// Efficiency of realtime [km/l]
		float			eff_t;			// Efficiency of trip mean [km/l]
		float			eff_s;			// Efficiency of all mean [km/l]

		unsigned char	x;				// Display position [dot]
		unsigned char	y;				// 
		unsigned char	w;				// Display size [dot]
		unsigned char	h;				//

		unsigned char	ect_color;		// 
		unsigned char	rpm_color;		// 
		unsigned char	vss_color;		// 
		unsigned char	vpwr_color;		// 

		//---------------------------------------------------------------------
		//	Input
		//---------------------------------------------------------------------

		//
		// Default value
		//
		ect       = 0;
		shrtft1   = 0;
		longft1   = 0;
		map       = 0;
		rpm       = 0;
		vss       = 0;
		sparkadv  = 0;
		iat       = 0;
		tp        = 0;
		o2s11     = 0;
		shrtft11  = 0;
		dist      = 0;
		fuse      = 0;
//		vpwr      = 0;


		//
		// Communication channel
		//
		if (collect) {
#define STAT_DISP_X			32
#define STAT_DISP_Y			(LCD_SCREEN_HEIGHT - LCD_FONT_HEIGHT)
#define STAT_COLOR_ERR		LCD_COLOR_RED
#define STAT_COLOR_WARN		LCD_COLOR_YELLOW
#define STAT_COLOR_INFO		LCD_COLOR_GREEN
			char	msg[32];

			if (line_init == 0) {		// Initiate ISO9141-2
				unsigned char	rc;
				unsigned char	kw[2];

				if ((rc = iso9141_init(kw)) != 0) {
					sprintf_P(msg, PSTR("ERR:ComInit=%d"), rc);
					lcd_draw_str(STAT_DISP_X, STAT_DISP_Y, STAT_COLOR_ERR, 0, msg);
					time_out++;
				} else {
					lcd_draw_str(STAT_DISP_X, STAT_DISP_Y, STAT_COLOR_INFO, 0, "OK Connected");
					line_init = 1;
					time_out = 0;
				}
			}

			while (line_init) {			// Query vhiecle information
				const unsigned char	pid[] = {0x05, 0x06, 0x07, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x11, 0x14};
				unsigned char	i;
				unsigned char	rc;
				unsigned char	buf[8];
				unsigned char	len;

				for(i = 0; i < sizeof(pid); i++) {
			
					buf[0] = 0x01;		// Request current powertrain diagnostic data request SID
					buf[1] = pid[i];
					if ((rc = iso9141_send_msg(buf, 2)) != 0) {
						break;
					}
					if ((rc = iso9141_recv_response(buf, &len)) != 0) {
						break;
					}
			
					switch(i){
						case 0:
							ect = buf[2] - 40;
							break;
						case 1:
							shrtft1 = (buf[2] - 128) * 100 / 128;
							break;
						case 2:
							longft1 = (buf[2] - 128) * 100 / 128;
							break;
						case 3:
							map = buf[2];
							break;
						case 4:
							rpm = (buf[2] * 256 + buf[3]) / 4;
							break;
						case 5:
							vss = buf[2];
							break;
						case 6:
							sparkadv = (buf[2] - 128) / 2;
							break;
						case 7:
							iat = buf[2] - 40;
							break;
						case 8:
							tp = buf[2] * 100 / 255;
							break;
						case 9:
							o2s11 = buf[2] * 5;
							shrtft11 = (buf[3] - 128) * 100 / 128;
							break;
						default:
							break;
					}
			
				}

				if (rc != 0) {
					sprintf_P(msg, PSTR("ERR:CommR/W=%d"), rc);
					lcd_draw_str(STAT_DISP_X, STAT_DISP_Y, STAT_COLOR_ERR, 0, msg);
					line_init = 0;
					break;
				}

				break;
			}
		}

		//
		// Battery voltage
		//
		vpwr = (long)getadc(0) * 3300L * (long)(RDIV_1 + RDIV_2)
		     / (long)(1024L * RDIV_2);


		//---------------------------------------------------------------------
		//	Action
		//---------------------------------------------------------------------

		//
		// Power off/on action
		//
		if (time_out >= 2) {
			unsigned char	b;

			// SDcard checkpoint
			LED_1_ON();
			sd_sync(fd);
			LED_1_OFF();

			// Power off
			LCD_BL_OFF();
			lcd_close();
			BEEP_TERM();

			// Wait at wakeup event
			b = 0;
			while(!pushkey) {
				unsigned char	v;

				sleep_mode();			// Break by interrupt (RTC 1Hz and Push-button)

//				v = getadc(0) >> 2;		// DEL 2012.07.07
				v = getadc(0) >> 4;		// ADD 2012.07.07
				if (b != 0) {
					if (b < v) {
						stoptm = 0;		// ADD 2012.07.07
						break;
					}
				}
				b = v;
			}

			// Power on
			BEEP_INIT();
			LCD_BL_ON();
			lcd_open();

			// Resume
			disp_mode_old = 0;			// Force redraw request

		}

		//
		// Push key
		//

		// Long push
		if (SWITCH_1() || SWITCH_2() || SWITCH_3() || SWITCH_4()) {
			pushkey_c++;
			if (pushkey_c >= 3) {
				if (SWITCH_1()) {
					pushkey = 5;
				}
				if (SWITCH_2()) {
					pushkey = 6;
				}
				if (SWITCH_3()) {
					pushkey = 7;
				}
				if (SWITCH_4()) {
					pushkey = 8;
				}
			}
		} else {
			pushkey_c = 0;
		}

		// One time action
		if (pushkey) {
			beep_approve();
			if (pushkey == 8) {
				setup();
			} else {
				disp_mode = pushkey;
			}
		}

		//
		// Calculate
		//
		atm = map * 1000.0 / 101325.0;					// Air pressure [atm]
		r = (1.293 * atm) / (1 + 0.00367 * iat);		// Air density [kg/m^3]
		v = (float)rpm / 60.0 / E_CYC * C_VOL;			// Air volume [m^3/sec]
		maf = r * v * 1000.0;							// Air flow on intake manifold [g/sec]

//		mff = maf / 14.7 * (shrtft1 / 100.0 + 1.0);		// DEL 2012.07.11  Fuel flow on intake manifold [g/sec]
		mff = maf / 14.7 * (shrtft11 / 100.0 + 1.0);	// ADD 2012.07.11  Fuel flow on intake manifold [g/sec]
		
		if (ss > ss_old) {								// Interval time of measure [s]
			intvl = ss - ss_old;						//
		} else {										//
			intvl = 60 - ss_old + ss;					//
		}												//
		ss_old = ss;									//

		dist = (vss / (60.0 * 60.0) * intvl);			// Interval distance [km]

		fuse = (mff / 0.783 / 1000.0 * intvl);			// Interval used fuel [l]

		// Estimate
		if (dd_old != dd) {								// Daily reset
			dd_old  = dd;								//
			dist_t  = 0;								//
			fuse_t  = 0;								//
			runtm_t = 0;								//
			ontm_t  = 0;								//
		}												//
		dist_t += dist;									// 
		dist_s += dist;									// 
		fuse_t += fuse;									//
		fuse_s += fuse;									//
		if (vss > 0) {									//
			runtm_t += intvl;							//
			runtm_s += intvl;							//
			stoptm   = 0;								//
		} else {										//
			stoptm += intvl;							//
		}												//
		if (rpm > 0) {									//
			ontm_t += intvl;							//
			ontm_s += intvl;							//
		}												//

		// Eefficient
		if (dist < 0.0001) {							// 10[cm]
			eff = 0.0;									//
		} else if (fuse > 0.00001) {					// 0.01[ml]
			eff = dist / fuse;							// [km/l]
		} else {										//
			eff = 99.9;									//
		}												//
		if (dist_t < 0.001) {							// 1[m]
			eff_t = 0.0;								//
		} else if (fuse_t > 0.001) {					// 1[ml]
			eff_t = dist_t / fuse_t;					//
		} else {										//
			eff_t = 99.9;								//
		}												//
		if (dist_s < 0.001) {							// 1[m]
			eff_s = 0.0;								//
		} else if (fuse_s > 0.001) {					// 1[ml]
			eff_s = dist_s / fuse_s;					//
		} else {										//
			eff_s = 99.9;								//
		}												//


		//---------------------------------------------------------------------
		//	Output
		//---------------------------------------------------------------------

		//
		// SDcard write
		//
		LED_1_ON();
		if (log_write) {
			if (CD()) {					// Insert check
				static short			yy_b = 0;
				static unsigned char	mm_b = 0;
				static unsigned char	dd_b = 0;

				if ((yy_b != yy) ||		// Change logfile
				    (mm_b != mm) ||
					(dd_b != dd)) {
					yy_b = yy;
					mm_b = mm;
					dd_b = dd;
					if (fd >= 0) {
						sd_close(fd);
						sd_unmount();
						fd = -1;
					}
				}
				
				if (fd < 0) {			// Open
					if (WP()) {
						lcd_draw_str(STAT_DISP_X, STAT_DISP_Y, STAT_COLOR_ERR, 0, "SDcard protect.");
					} else {
						if (sd_mount(0) < 0) {
							lcd_draw_str(STAT_DISP_X, STAT_DISP_Y, STAT_COLOR_ERR, 0, "SDcard media err.");
						} else {
							char	fname[16];

							sprintf_P(fname, PSTR("%04d%02d%02d.CSV"), yy, mm, dd);
							SD_SET_DATE(yy, mm, dd);
							SD_SET_TIME(hh, mi, ss);
							if ((fd = sd_open(fname, SD_O_CREAT|SD_O_APPEND, 0)) < 0) {
								lcd_draw_str(STAT_DISP_X, STAT_DISP_Y, STAT_COLOR_ERR, 0, "SDcard file err.");
								sd_unmount();
							} else {	// CSV header
								printf_P(PSTR("\"DateTime\",\"ect[`C]\",\"shrtft1[%%]\",\"longft1[%%]\",\"map[kPa]\",\"rpm[rpm]\",\"vss[km/h]\",\"sparkadv[`]\",\"iat[`C]\",\"tp[%%]\",\"o2s11[mV]\",\"shrtft11[%%]\",\"vpwr[mV]\",\"maf[g/s]\",\"mff[g/s]\",\"dist[m]\",\"fuse[ml]\"\r\n"));
							}
						}
					}
				}

				if (fd >= 0) {			// Append write
					printf_P(PSTR("\"%d-%02d-%02d %02d:%02d:%02d\","), yy, mm, dd, hh, mi, ss);
					printf_P(PSTR("%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%u,%d.%d,%d.%d,%d.%d,%d.%d\r\n"),
						ect, shrtft1, longft1, map, rpm, vss, sparkadv, iat, tp, o2s11, shrtft11, vpwr,
						(int)maf, (int)(maf*10.0)%10,
						(int)mff, (int)(mff*10.0)%10,
						(int)(dist*1000.0), (int)(dist*10000.0)%10,
						(int)(fuse*1000.0), (int)(fuse*10000.0)%10);
					if (stoptm > 0) {
						sd_sync(fd);
					}
				}
			}
		} else {
			if (fd >= 0) {
				sd_close(fd);
				sd_unmount();
				fd = -1;
			}
		}
		LED_1_OFF();

		//
		// 	Beep warning
		//
		if (ect < 50) {
			ect_color = LCD_COLOR_BLUE;
		} else if (ect > 90) {
			ect_color = LCD_COLOR_RED;
			beep_warn();
		} else {
			ect_color = LCD_COLOR_GREEN;
		}
		if (rpm < 3000) {
			rpm_color = LCD_COLOR_GREEN;
		} else if (rpm < 5000) {
			beep_warn();
		} else {
			rpm_color = LCD_COLOR_RED;
			beep_warn();
		}
		if (vss < 50) {
			vss_color = LCD_COLOR_GREEN;
		} else if (vss < 80) {
			vss_color = LCD_COLOR_YELLOW;
		} else {
			vss_color = LCD_COLOR_RED;
			beep_warn();
		}
		if ((vpwr > 12000) && (vpwr < 15000)) {
			vpwr_color = LCD_COLOR_GREEN;
		} else if ((vpwr > 11500) && (vpwr < 15500)) {
			vpwr_color = LCD_COLOR_YELLOW;
		} else {
			vpwr_color = LCD_COLOR_RED;
			beep_warn();
		}


		//
		// Auto brighteness
		//
		LCD_BL_SET(MIN(((1023 - getadc(2)) >> 2) + 10, 0xff));	// ADD 2012.07.08  minimum 8/256=3%


		//
		// Function display
		//
		if (disp_mode != disp_mode_old) {
			unsigned char	i;
			
			y = FB_START_Y;
			for(i = 0; i < 5; i++) {
				lcd_draw_line(FB_START_X, y, ((FB_CELL_W + FB_BORDER_SIZE) * 1 + FB_START_X), y, 1, FB_BORDER_COLOR);
				y += (FB_BORDER_SIZE + FB_CELL_H);
			}
			x = FB_START_X;
			for(i = 0; i < 2; i++) {
				lcd_draw_line(x, FB_START_Y, x, ((FB_CELL_H + FB_BORDER_SIZE) * 4 + FB_START_Y), 1, FB_BORDER_COLOR);
				x += (FB_BORDER_SIZE + FB_CELL_W);
			}

			for(i = 1; i <= 4; i++) {
				FB_DRAW_CELL(i, FB_CHAR_FCOLOR, FB_CHAR_BCOLOR, fb_msg[i]);
			}

			x = FB_APPL_X;
			y = FB_APPL_Y;
			w = FB_APPL_W;
			h = FB_APPL_H;
			lcd_draw_rect(x,y, w,h, 0);	// Clear application area

			if (disp_mode > 0) {
//				FB_DRAW_CELL(disp_mode%4, FB_CHAR_BCOLOR, FB_CHAR_SCOLOR, fb_msg[disp_mode]);	// DEL 2012.07.08
				FB_DRAW_CELL(disp_mode%5, FB_CHAR_BCOLOR, FB_CHAR_SCOLOR, fb_msg[disp_mode]);	// ADD 2012.07.08
			}
		}

		//
		// Efficiency display
		//
		if (disp_mode == 1) {
			char			msg[60];
			float			t1;
			float			t2;
			unsigned char	cc;

			// Fuel efficiency
			if (eff > eff_s) {
				cc = LCD_COLOR_GREEN;
			} else {
				cc = LCD_COLOR_YELLOW;
			}
			v_gauge(FB_APPL_X, FB_APPL_Y, (int)(FB_APPL_W*0.7), 108,
			        cc, LCD_COLOR_WHITE, LCD_COLOR_GRAY, LCD_COLOR_BLACK,
//			        eff, eff_s, 20.0, 4);		// DEL 2012.07.07
			        eff, eff_s, 30.0, 3);		// ADD 2012.07.07
			lcd_draw_str(FB_APPL_X, FB_APPL_Y+10, LCD_COLOR_GRAY, LCD_COLOR_BLACK, "km/l");
			sprintf_P(msg, PSTR("Ave%3d.%dkm/l\nDay%3d.%dkm/l"),
				(int)eff_s, ((int)(eff_s*10.0))%10,
				(int)eff_t, ((int)(eff_t*10.0))%10);
			lcd_draw_str(FB_APPL_X, FB_APPL_Y+108, LCD_COLOR_GREEN, LCD_COLOR_BLACK, msg);

			// Running time rate
			t1 = (ontm_t == 0) ? 0.0 : (float)runtm_t / (float)ontm_t * 100.0;
			t2 = (ontm_s == 0) ? 0.0 : (float)runtm_s / (float)ontm_s * 100.0;
			v_gauge(FB_APPL_X+(int)(FB_APPL_W*0.7), FB_APPL_Y, (int)(FB_APPL_W*0.3), 108,
			        LCD_COLOR_GREEN, LCD_COLOR_YELLOW, LCD_COLOR_GRAY, LCD_COLOR_BLACK,
			        t1, t2, 100.0, 5);
			lcd_draw_str(FB_APPL_X+(int)(FB_APPL_W*0.7), FB_APPL_Y+10, LCD_COLOR_GRAY, LCD_COLOR_BLACK, "Run%");

			sprintf_P(msg, PSTR("%2d:%02d\n%2d:%02d"),
				(int)(ontm_t / 3600), (int)((ontm_t / 60) % 60),
				(stoptm / 60), (stoptm % 60));
			lcd_draw_str(FB_APPL_X+(int)(FB_APPL_W*0.7), FB_APPL_Y+108, LCD_COLOR_GREEN, LCD_COLOR_BLACK, msg);
		}

		//
		// Efficiency and Speed display
		//
		if (disp_mode == 2) {
//#define EFF_MAX			20			// DEL 2012.07.07  [km/l]
#define EFF_MAX				30			// ADD 2012.07.07  [km/l]
#define VSS_MAX				80			// [km/h]
#define EFF_DIV				3			// ADD 2012.07.07  Guage scale
#define VSS_DIV				4			// ADD 2012.07.07  Guage scale
#define EFF_VSS_ORIGIN_X	(FB_APPL_X + 24)
#define EFF_VSS_ORIGIN_Y	(FB_APPL_Y + 108)
#define EFF_VSS_GRAPH_W		(80)
#define EFF_VSS_GRAPH_H		(100)
#define EFF_VSS_DIV_Y		3
#define EFF_VSS_DIV_X		4
			unsigned char	i;

			if (disp_mode != disp_mode_old) {
				// Scale line
				for(i = 0; i <= EFF_VSS_DIV_Y; i++) {
					unsigned char	y1;

					y1 = EFF_VSS_ORIGIN_Y - ((EFF_VSS_GRAPH_H * i) / EFF_VSS_DIV_Y);
					lcd_draw_line(EFF_VSS_ORIGIN_X, y1, (EFF_VSS_ORIGIN_X + EFF_VSS_GRAPH_W), y1, 1, LCD_COLOR_GRAY);
				}
				for(i = 0; i <= EFF_VSS_DIV_X; i++) {
					unsigned char	x1;

					x1 = EFF_VSS_ORIGIN_X + ((EFF_VSS_GRAPH_W * i) / EFF_VSS_DIV_X);
					lcd_draw_line(x1, EFF_VSS_ORIGIN_Y, x1, (EFF_VSS_ORIGIN_Y - EFF_VSS_GRAPH_H), 1, LCD_COLOR_GRAY);
				}
				lcd_draw_str((EFF_VSS_ORIGIN_X + 1), (EFF_VSS_ORIGIN_Y - EFF_VSS_GRAPH_H + 2), LCD_COLOR_GRAY, LCD_COLOR_BLACK, "km/l");
				lcd_draw_str((EFF_VSS_ORIGIN_X + EFF_VSS_GRAPH_W - 25) , (EFF_VSS_ORIGIN_Y - 9), LCD_COLOR_GRAY, LCD_COLOR_BLACK, "km/h");
			}

			// Bar graph
			v_gauge(FB_APPL_X, (EFF_VSS_ORIGIN_Y - EFF_VSS_GRAPH_H),
					(EFF_VSS_ORIGIN_X - FB_APPL_X - 1), EFF_VSS_GRAPH_H,
			        LCD_COLOR_GREEN, LCD_COLOR_WHITE, LCD_COLOR_GRAY, LCD_COLOR_BLACK,
//			        eff, eff, EFF_MAX, 4);			// DEL 2012.07.07
			        eff, eff, EFF_MAX, EFF_DIV);	// ADD 2012.07.07

			h_gauge(EFF_VSS_ORIGIN_X, EFF_VSS_ORIGIN_Y+1,
					EFF_VSS_GRAPH_W, (FB_APPL_H - EFF_VSS_ORIGIN_Y),
			        vss_color, LCD_COLOR_WHITE, LCD_COLOR_GRAY, LCD_COLOR_BLACK,
//			        (float)vss, (float)vss, VSS_MAX, 4);		// DEL 2012.07.07
			        (float)vss, (float)vss, VSS_MAX, VSS_DIV);	// ADD 2012.07.07

			// Dot point
			{
				unsigned char	x1;
				unsigned char	y1;

				x1 = EFF_VSS_ORIGIN_X + (vss * EFF_VSS_GRAPH_W) / VSS_MAX;
				y1 = EFF_VSS_ORIGIN_Y - (rpm * EFF_VSS_GRAPH_H) / EFF_MAX;
				lcd_draw_rect(x1, y1, 2, 2, LCD_COLOR_WHITE);
			}
		}

		//
		// RPM and Speed display
		//
		if (disp_mode == 3) {
#define RPM_MAX				6000		// [krpm]
//#define VSS_MAX			80			// [km/h]
#define RPM_VSS_ORIGIN_X	(FB_APPL_X + 24)
#define RPM_VSS_ORIGIN_Y	(FB_APPL_Y + 108)
#define RPM_VSS_GRAPH_W		(80)
#define RPM_VSS_GRAPH_H		(100)
#define RPM_VSS_DIV_Y		3
#define RPM_VSS_DIV_X		4
			unsigned char	i;

			if (disp_mode != disp_mode_old) {
				// Scale line
				for(i = 0; i <= RPM_VSS_DIV_Y; i++) {
					unsigned char	y1;

					y1 = RPM_VSS_ORIGIN_Y - ((RPM_VSS_GRAPH_H * i) / RPM_VSS_DIV_Y);
					lcd_draw_line(RPM_VSS_ORIGIN_X, y1, (RPM_VSS_ORIGIN_X + RPM_VSS_GRAPH_W), y1, 1, LCD_COLOR_GRAY);
				}
				for(i = 0; i <= RPM_VSS_DIV_X; i++) {
					unsigned char	x1;

					x1 = RPM_VSS_ORIGIN_X + ((RPM_VSS_GRAPH_W * i) / RPM_VSS_DIV_X);
					lcd_draw_line(x1, RPM_VSS_ORIGIN_Y, x1, (RPM_VSS_ORIGIN_Y - RPM_VSS_GRAPH_H), 1, LCD_COLOR_GRAY);
				}
				lcd_draw_str((RPM_VSS_ORIGIN_X + 1), (RPM_VSS_ORIGIN_Y - RPM_VSS_GRAPH_H + 2), LCD_COLOR_GRAY, LCD_COLOR_BLACK, "krpm");
				lcd_draw_str((RPM_VSS_ORIGIN_X + RPM_VSS_GRAPH_W - 25) , (RPM_VSS_ORIGIN_Y - 9), LCD_COLOR_GRAY, LCD_COLOR_BLACK, "km/h");
			}

			// Bar graph
			v_gauge(FB_APPL_X, (RPM_VSS_ORIGIN_Y - RPM_VSS_GRAPH_H),
					(RPM_VSS_ORIGIN_X - FB_APPL_X - 1), RPM_VSS_GRAPH_H,
			        rpm_color, LCD_COLOR_WHITE, LCD_COLOR_GRAY, LCD_COLOR_BLACK,
			        rpm/1000.0, rpm/1000.0, RPM_MAX/1000.0, 3);

			h_gauge(RPM_VSS_ORIGIN_X, RPM_VSS_ORIGIN_Y+1,
					RPM_VSS_GRAPH_W, (FB_APPL_H - RPM_VSS_ORIGIN_Y),
			        vss_color, LCD_COLOR_WHITE, LCD_COLOR_GRAY, LCD_COLOR_BLACK,
			        (float)vss, (float)vss, VSS_MAX, 4);

			// Dot point
			{
				unsigned char	x1;
				unsigned char	y1;

				x1 = RPM_VSS_ORIGIN_X + (vss * RPM_VSS_GRAPH_W) / VSS_MAX;
				y1 = RPM_VSS_ORIGIN_Y - (rpm * RPM_VSS_GRAPH_H) / RPM_MAX;
				lcd_draw_rect(x1, y1, 2, 2, LCD_COLOR_WHITE);
			}
		}


		//
		// Sub panel display
		//
		if (disp_mode == 4) {
			v_gauge(FB_APPL_X, FB_APPL_Y, (int)(FB_APPL_W*0.25), (int)(FB_APPL_H*0.9),
			        ect_color, LCD_COLOR_WHITE, LCD_COLOR_GRAY, LCD_COLOR_BLACK,
			        ect, ect, 100.0, 5);
			lcd_draw_str(FB_APPL_X, FB_APPL_Y+10, LCD_COLOR_GRAY, 0, "\177C");
			lcd_draw_str(FB_APPL_X+(int)(FB_APPL_W*0.05), FB_APPL_Y+(int)(FB_APPL_H*0.9), LCD_COLOR_GREEN, LCD_COLOR_BLACK, "ENG");

			v_gauge(FB_APPL_X+(int)(FB_APPL_W*0.25), FB_APPL_Y, (int)(FB_APPL_W*0.5), (int)(FB_APPL_H*0.9),
			        rpm_color, LCD_COLOR_WHITE, LCD_COLOR_GRAY, LCD_COLOR_BLACK,
			        rpm/1000.0, rpm/1000.0, 6.0, 6);
			lcd_draw_str(FB_APPL_X+(int)(FB_APPL_W*0.25), FB_APPL_Y+10, LCD_COLOR_GRAY, LCD_COLOR_BLACK, "krpm");
			lcd_draw_str(FB_APPL_X+(int)(FB_APPL_W*0.45), FB_APPL_Y+(int)(FB_APPL_H*0.9), LCD_COLOR_GREEN, 0, "RPM");

			v_gauge(FB_APPL_X+(int)(FB_APPL_W*0.75), FB_APPL_Y, (int)(FB_APPL_W*0.25), (int)(FB_APPL_H*0.9),
			        vpwr_color, LCD_COLOR_WHITE, LCD_COLOR_GRAY, LCD_COLOR_BLACK,
			        vpwr/1000.0, vpwr/1000.0, 18.0, 3);
			lcd_draw_str(FB_APPL_X+(int)(FB_APPL_W*0.75), FB_APPL_Y+10, LCD_COLOR_GRAY, LCD_COLOR_BLACK, "V");
			lcd_draw_str(FB_APPL_X+(int)(FB_APPL_W*0.75), FB_APPL_Y+(int)(FB_APPL_H*0.9), LCD_COLOR_GREEN, 0, "BATT");

		}


		//
		// Effciency numerical value
		//
		if (disp_mode == 5) {
			unsigned char	i;
			unsigned char	x1;
			unsigned char	y1;

			x1 = x + 3;
			y1 = y + 4;

			for(i = 0; i < 11; i++) {
				char		msg[32];

				switch(i) {
				case 0:
					sprintf_P(msg, PSTR("     Total  Day"));
					break;
				case 1:
					sprintf_P(msg, PSTR("Dist%4d.%d%3d.%dkm"),
						(int)dist_s, ((int)(dist_s*10.0))%10,
						(int)dist_t, ((int)(dist_t*10.0))%10);
					break;
				case 2:
					sprintf_P(msg, PSTR("Fuel%4d.%d%3d.%dl"),
						(int)fuse_s, ((int)(fuse_s*10.0))%10,
						(int)fuse_t, ((int)(fuse_t*10.0))%10);
					break;
				case 3:
					sprintf_P(msg, PSTR("Effc%4d.%d%3d.%d"),
						(int)eff_s, ((int)(eff_s*10.0))%10,
						(int)eff_t, ((int)(eff_t*10.0))%10);
					break;
				case 4:
					sprintf_P(msg, PSTR("Run%4d:%02d%3d:%02d"),
						(int)(runtm_s/3600), (int)((runtm_s/60)%60),
						(int)(runtm_t/3600), (int)((runtm_t/60)%60));
					break;
				case 5:
					sprintf_P(msg, PSTR("On %4d:%02d%3d:%02d"),
						(int)(ontm_s/3600), (int)((ontm_s/60)%60),
						(int)(ontm_t/3600), (int)((ontm_t/60)%60));
					break;
				case 6:
					sprintf_P(msg, PSTR("Run%%%6d%6d"),
						(int)(runtm_s * 100 / ontm_s),
						(int)(runtm_t * 100 / ontm_t));
					break;
				case 7:
					sprintf_P(msg, PSTR("-----------------"));
					break;
				case 8:
					sprintf_P(msg, PSTR("AirFlow%4d.%dg/s"),
						(int)maf,
						(int)(maf*10.0)%10);
					break;
				case 9:
					sprintf_P(msg, PSTR("FuelFlow%3d.%dg/s"),
						(int)mff,
						(int)(mff*10.0)%10);
					break;
				case 10:
					sprintf_P(msg, PSTR("Efficien%3d.%dkm/l"),
						(int)eff,
						(int)(eff*10.0)%10);
					break;
				default:
					break;
				}

				lcd_draw_str(x1, y1, LCD_COLOR_GREEN, LCD_COLOR_BLACK, msg);
				y1 += 10;

			}
		}

		//
		// Detail display
		//
		if (disp_mode == 6) {
			unsigned char	i;
			unsigned char	x1;
			unsigned char	y1;

			x1 = x + 3;
			y1 = y + 4;

			for(i = 0; i < 12; i++) {
				prog_char	*p;
				int			v;
				char		msg[32];

				switch(i) {
				case 0:
					p = PSTR("ECT %8d \177C");
					v = ect;
					break;
				case 1:
					p = PSTR("SHRTFT1 %4d %%");
					v = shrtft1;
					break;
				case 2:
					p = PSTR("LONGFT1 %4d %%");
					v = longft1;
					break;
				case 3:
					p = PSTR("MAP %8d kPa");
					v = map;
					break;
				case 4:
					p = PSTR("RPM %8d rpm");
					v = rpm;
					break;
				case 5:
					p = PSTR("VSS %8d km/h");
					v = vss;
					break;
				case 6:
					p = PSTR("SPARKADV %3d \177");
					v = sparkadv;
					break;
				case 7:
					p = PSTR("IAT %8d \177C");
					v = iat;
					break;
				case 8:
					p = PSTR("TP %9d %%");
					v = tp;
					break;
				case 9:
					p = PSTR("O2S11 %6d mV");
					v = o2s11;
					break;
				case 10:
					p = PSTR("SHRTFT11 %3d %%");
					v = shrtft11;
					break;
				case 11:
					p = PSTR("VPWR %7d mV");
					v = vpwr;
					break;
				default:
					p = PSTR("");
					v = 0;
					break;
				}

				sprintf_P(msg, p, v);
				lcd_draw_str(x1, y1, LCD_COLOR_GREEN, 0, msg);
				y1 += 10;

			}
		}

		//---------------------------------------------------------------------
		//	Next interval
		//---------------------------------------------------------------------

		pushkey = 0;
		disp_mode_old = disp_mode;


	}

	return(0);
}
