//
//  iso9141_io.c
//  ISO9141-2 and ISO15031-5(SAE J1979) protocol simple driver
//  Copyright(C)2012 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2012.02.27 v0.0  First cording
//

#include <avr/io.h>						// WinAVR library
#include <util/delay.h>

#include "uart.h"
#include "iso9141_io.h"


unsigned char iso9141_open(void){
	return(uart_open());
}


unsigned char iso9141_close(void){
	return(uart_close());
}


unsigned char iso9141_init(unsigned char keyword[2]){
	unsigned char	rc;
	unsigned char	addr[] = {0x33};

	do {
		unsigned char	i;
		unsigned char	c;
		unsigned char	l;
		
		// 5 baud method
		_delay_ms(W0_MIN);
		if (uart_send_slow(addr, sizeof(addr)) != 0) {
			rc = ISO9141_ERR_SEND;
			break;
		}
		
		// Recieve sync
		_delay_ms(W1_MIN);
		for(i = 0; i < SYNC_TRY; i++) {
			if ((l = uart_recv(&c, 1)) == 0) {
				if (c == 0x55) {
					break;
				}
			}
		}
		if (i >= SYNC_TRY) {
			rc = ISO9141_ERR_TIMEOUT_SYNC;
			break;
		}
		
		// Recieve key word bytes
		if (uart_recv(keyword, sizeof(keyword)) != 0) {
			rc = ISO9141_ERR_TIMEOUT_KEYWORD;
			break;
		}
		
		// Send key word 2
		_delay_ms(W4_MIN);
		c = ~keyword[1];
		if (uart_send(&c, 1) != 0) {
			rc = ISO9141_ERR_SEND;
			break;
		}
		
		// Recieve my address
		if (uart_recv(&c, 1) != 0) {
			rc = ISO9141_ERR_TIMEOUT_ADDR;
			break;
		}
		if (c != (unsigned char)(~addr[0])) {	// Unmatch
			rc = ISO9141_ERR_PROTOCOL_ADDR;	
			break;
		}
		
		rc = 0;
		break;
		
	} while(1);
	
	return(rc);
};


#define H1	0x68
#define H2	0x6a
#define H3	0xf1

unsigned char iso9141_send_msg(unsigned char *data, unsigned char len){
	unsigned char	rc;
	unsigned char	head[3] = {H1, H2, H3};
	unsigned char	csum[1];
	int				s;
	unsigned char	i;
	
	
	// Calc check sum
	s = H1 + H2 + H3;
	for(i = 0; i < len; i++) {
		s += data[i];
	}
	csum[0] = s;
		
	// Send 3 part timming
	do {
		rc = 0;

		_delay_ms(P3_MIN);

		if (uart_send(head, sizeof(head)) != 0) {
			rc = ISO9141_ERR_SEND;
			break;
		}

		_delay_ms(P4_MIN);

		if (uart_send(data, len) != 0) {
			rc = ISO9141_ERR_SEND;
			break;
		}

		_delay_ms(P4_MIN);

		if (uart_send(csum, sizeof(csum)) != 0) {
			rc = ISO9141_ERR_SEND;
			break;
		}

		break;

	} while(1);
	
	return(rc);
}


unsigned char iso9141_recv_response(unsigned char *data, unsigned char *len){
	unsigned char	rc;
	unsigned char	llen;
	
	llen = 0;
	do {
		unsigned char	buf[11];		// Max response message size
		unsigned char	l;
		unsigned char	i;
		
		// Recv
		if ((rc = uart_recv(buf, sizeof(buf))) == sizeof(buf)) {
			rc = ISO9141_ERR_TIMEOUT_RESPONSE;
			break;
		}
		l = sizeof(buf) - rc;
				
		// Sum check
		if (l > 1) {
			int	s;

			s = 0;
			--l;
			for(i = 0; i < l; i++) {
				s += buf[i];
			}
			if ((unsigned char)s != buf[l]) {
				rc = ISO9141_ERR_PROTOCOL_CSUM;
				break;
			}
		}
		
		// Set response data
		for(i = 3; i < l; i++) {
			*data++ = buf[i];
			llen++;
		}

		rc = 0;
		break;
		
	} while(1);

	*len = llen;	
	return(rc);
}

