//
//  l1f10085.h
//  Graphic LCD module "EPSON L1F10085" I/O library
//  Copyright(C)2012 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//	Description
//		I/O control for Graphic LCD module.
//		LCD module: EPSON L1F10085. (Controller chip is EPSON S1D15G10)
//		Interface: 9bits SPI by software control.
//			MOSI : Any output port (LCD_PIN_SDATA)
//			MISO : Not used
//			SCLK : Any output port (LCD_PIN_SCK)
//			CS   : Any output port (LCD_PIN_CS)
//			RESET: Any output port (LCD_PIN_RESET) ...optinal
//		Screen direction:
//			X: 0=Left .. 129=Right
//			Y: 0=Top .. 129=Bottom
//		Color mode: 256 index
//
//  Device: ATmega328P       IDE: AVR Studio 4.18.700, WinAVR 20100110(GCC 4.3.3)
//
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2012.02.20 v0.0  First cording
//


//------------------------------------------------------------------------------
// Hardware config & Option
//------------------------------------------------------------------------------
#define LCD_PIN_RESET_ON()		(PORTD &= ~_BV(PORTD0))							// Option Negative
#define LCD_PIN_RESET_OFF()		(PORTD |=  _BV(PORTD0))							// Option
#define LCD_PIN_RESET_INIT()	(DDRD  |=  _BV(DDD0), LCD_PIN_RESET_OFF())		// Option
#define LCD_PIN_RESET_TERM()	(DDRD  &= ~_BV(DDD0), PORTD &= ~_BV(PORTD0))	// Option

#define LCD_PIN_SDATA_ON()		(PORTD |=  _BV(PORTD1))
#define LCD_PIN_SDATA_OFF()		(PORTD &= ~_BV(PORTD1))
#define LCD_PIN_SDATA_INIT()	(DDRD  |=  _BV(DDD1), LCD_PIN_SDATA_OFF())
#define LCD_PIN_SDATA_TERM()	(DDRD  &= ~_BV(DDD1), PORTD &= ~_BV(PORTD1))

#define LCD_PIN_CS_ON()			(PORTD &= ~_BV(PORTD3))							// Negative
#define LCD_PIN_CS_OFF()		(PORTD |=  _BV(PORTD3))
#define LCD_PIN_CS_INIT()		(DDRD  |=  _BV(DDD3), LCD_PIN_CS_OFF())
#define LCD_PIN_CS_TERM()		(DDRD  &= ~_BV(DDD3), PORTD &= ~_BV(PORTD3))

#define LCD_PIN_SCK_ON()		(PORTD |=  _BV(PORTD4))
#define LCD_PIN_SCK_OFF()		(PORTD &= ~_BV(PORTD4))
#define LCD_PIN_SCK_INIT()		(DDRD  |=  _BV(DDD4), LCD_PIN_SCK_OFF())
#define LCD_PIN_SCK_TERM()		(DDRD  &= ~_BV(DDD4), PORTD &= ~_BV(PORTD4))

//#define LCD_VCD						// Enable VCD(Virtual Charactor Display)
//#define F_CPU 8000000UL				// Please IDE project configuration option frequency[Hz]


//------------------------------------------------------------------------------
// Layer 1
//------------------------------------------------------------------------------

extern void lcd_start(void);
extern void lcd_init(void);
extern void lcd_stop(void);
extern void lcd_sendCmdData(unsigned char data_or_command, unsigned char data) ;
#define LCD_DC_COMMAND		0
#define LCD_DC_DATA			1
// LCD module native command code
#define LCD_CMD_DISON		0xaf		// Display on                         None 
#define LCD_CMD_DISOFF		0xae		// Display off                        None 
#define LCD_CMD_DISNOR		0xa6		// Normal display                     None 
#define LCD_CMD_DISINV		0xa7		// Inverse display                    None 
#define LCD_CMD_COMSCN		0xbb		// Common scan direction              1byte 
#define LCD_CMD_DISCTL		0xca		// Display control                    3byte 
#define LCD_CMD_SLPIN		0x95		// Sleep in                           None 
#define LCD_CMD_SLPOUT		0x94		// Sleep out                          None 
#define LCD_CMD_PASET		0x75		// Page address set                   2byte 
#define LCD_CMD_CASET		0x15		// Column address set                 2byte 
#define LCD_CMD_DATCTL		0xbc		// Data scan direction, etc.          3byte 
#define LCD_CMD_RGBSET8		0xce		// 256-color position set             20byte 
#define LCD_CMD_RAMWR		0x5c		// Writing to memory                  Data 
#define LCD_CMD_RAMRD		0x5d		// Reading from memory                Data 
#define LCD_CMD_PTLIN		0xa8		// Partial display in                 2byte 
#define LCD_CMD_PTLOUT		0xa9		// Partial display out                None 
#define LCD_CMD_RMWIN		0xe0		// Read and modify write              None 
#define LCD_CMD_RMWOUT		0xee		// End                                None 
#define LCD_CMD_ASCSET		0xaa		// Area scroll set                    4byte 
#define LCD_CMD_SCSTART		0xab		// Scroll start set                   1byte 
#define LCD_CMD_OSCON		0xd1		// Internal oscillation on            None 
#define LCD_CMD_OSCOFF		0xd2		// Internal oscillation off           None 
#define LCD_CMD_PWRCTR		0x20		// Power control                      1byte 
#define LCD_CMD_VOLCTR		0x81		// Electronic volume control          2byte 
#define LCD_CMD_VOLUP		0xd6		// Increment electronic control by 1  None 
#define LCD_CMD_VOLDOWN		0xd7		// Decrement electronic control by 1  None 
#define LCD_CMD_TMPGRD		0x82		// Temperature gradient set           14byte 


//------------------------------------------------------------------------------
//	Layer 2
//------------------------------------------------------------------------------

//
// Prototype
//
#define     lcd_open()		(lcd_start(), lcd_init())
#define     lcd_close()		lcd_stop()
extern void lcd_draw_rect(unsigned char x, unsigned char y, unsigned char width, unsigned char height, unsigned char color);
extern void lcd_draw_line(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2, unsigned char line_size, unsigned char color);
extern void lcd_draw_str(unsigned char x, unsigned char y, unsigned char face_color, unsigned char back_color, char *str);
#ifdef LCD_VCD
extern void lcd_setViewPort(unsigned char x, unsigned char y, unsigned char width, unsigned char height);
extern void lcd_prString(char *str);
#define     lcd_prCString(s)	lcd_prString(s)
extern void lcd_position(unsigned char column, unsigned char row);
extern void lcd_color(unsigned char face_color);
extern void lcd_colors(unsigned char face_color, unsigned char back_color);
extern void lcd_control(unsigned char command);
#endif

//
// General parameter
//
#define LCD_SCREEN_WIDTH	130
#define LCD_SCREEN_HEIGHT	130
#define LCD_FONT_WIDTH		6
#define LCD_FONT_HEIGHT		8

//
// Color code
//                            rrrgggbb
#define LCD_COLOR_RED		0b11100000
#define LCD_COLOR_PINK		0b11100001
#define LCD_COLOR_PURPLE	0b11100010
#define LCD_COLOR_MAGENTA	0b11100011
#define LCD_COLOR_BLUE		0b00000011
#define LCD_COLOR_CYAN		0b00011111
#define LCD_COLOR_GREEN		0b00011100
#define LCD_COLOR_LIME		0b01111100
#define LCD_COLOR_YELLOW	0b11111100
#define LCD_COLOR_ORANGE	0b11110000
#define LCD_COLOR_BLACK		0b00000000
#define LCD_COLOR_GRAY		0b01101110
#define LCD_COLOR_WHITE		0b11111111

//
//	VCD lcd_control() command code
//
#define LCD_DISP_ON                       0x0C
#define LCD_DISP_OFF                      0x08
#define LCD_DISP_BLANK                    0x0A
#define LCD_DISP_CLEAR_HOME               0x01
#define LCD_CURSOR_ON                     0x0E
#define LCD_CURSOR_OFF                    0x0C
#define LCD_CURSOR_WINK                   0x0D
#define LCD_CURSOR_BLINK                  0x0F
#define LCD_CURSOR_SH_LEFT                0x10
#define LCD_CURSOR_SH_RIGHT               0x14
#define LCD_CURSOR_HOME                   0x02
#define LCD_CURSOR_LEFT                   0x04
#define LCD_CURSOR_RIGHT                  0x06
#define LCD_PORT_MASK                     0x7F
