//
//  uart.c
//  Software UART
//  Copyright(C)2012 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2012.02.27 v0.0  First cording
//

#include <avr/io.h>						// WinAVR library
#include <avr/interrupt.h>				// 
#include <util/delay.h>

#include "uart.h"


unsigned char uart_open(void){
	unsigned char	rc;

	rc = 0;
	RX_INIT();
	TX_INIT();
	TX_ON();
	LED_1_INIT();
	return(rc);
}


unsigned char uart_close(void){
	unsigned char	rc;

	rc = 0;
	RX_TERM();
	TX_TERM();
	return(rc);
}


unsigned char uart_send(unsigned char *dat, unsigned char len){
	unsigned char	rc;
	unsigned char	s;

	rc = 0;
	for(; len > 0; --len) {
		unsigned char	c;
		unsigned char	i;
		
		c = *dat++;

		s = SREG;
		cli();
		
		// Start bit
		TX_OFF();
		LED_1_OFF();
		_delay_us(BIT_TIME);
		
		// Data bits
		for(i = 8; i > 0; --i) {
			if (c & 0x01) {
				TX_ON();
				LED_1_ON();
			} else {
				TX_OFF();
				LED_1_OFF();
			}
			c >>= 1;
			_delay_us(BIT_TIME - DELAY_OFFSET);
		}
		
		// Stop bit
		TX_ON();
		LED_1_ON();
		_delay_us(BIT_TIME);

//		if (s & _BV(SREG_I)) {				// Resume global interrupt
//			sei();
//		}
		SREG = s;

	}
	
	return(rc);
}

//
// Special slow version
//
unsigned char uart_send_slow(unsigned char *dat, unsigned char len){
	unsigned char	rc;
	
	rc = 0;
	for(; len > 0; --len) {
		unsigned char	c;
		unsigned char	i;
		
		c = *dat++;
		
		// Start bit
		TX_OFF();
		LED_1_OFF();
		_delay_ms(BIT_TIME_SLOW);
		
		// Data bits
		for(i = 8; i > 0; --i) {
			if (c & 0x01) {
				TX_ON();
				LED_1_ON();
			} else {
				TX_OFF();
				LED_1_OFF();
			}
			c = c >> 1;
			_delay_ms(BIT_TIME_SLOW);
		}
		
		// Stop bit
		TX_ON();
		LED_1_ON();
		_delay_ms(BIT_TIME_SLOW);
	}
	
	return(rc);
}


unsigned char uart_recv(unsigned char *dat, unsigned char len){
	unsigned char	s;

	for(; len > 0; --len) {
		unsigned short	i;
		unsigned char	j;
		unsigned char	c;

		// Wait high level (stop bit or idle)
		for(i = TIME_OUT_COUNT; i > 0 ; --i) {
			if (RX()) {
				LED_1_ON();
				break;
			}
			LED_1_OFF();
		}
		if (i == 0) {
			break;
		}

		// Wait low level (start bit)
		j = 0;
		for(i = TIME_OUT_COUNT; i > 0; --i) {
			if (!RX()) {
				if (!RX()) {
					LED_1_OFF();
					break;
				}
			}
			LED_1_ON();
		}
		if (i == 0) {
			break;
		}

		// Get data bits
		s = SREG;
		cli();
		c = 0;
		_delay_us((int)(BIT_TIME * 0.5));
		for(i = 8; i > 0; --i) {
			_delay_us(BIT_TIME - DELAY_OFFSET);
			c >>= 1;
			if (RX()) {
				c |= 0x80;
				LED_1_ON();
			} else {
				LED_1_OFF();
				LED_1_OFF();
			}
		}
//		if (s & _BV(SREG_I)) {				// Resume global interrupt
//			sei();
//		}
		SREG = s;

		// Save data
		*dat++ = c;
	}

	return(len);						// Free dat size
}
